/* 
 *  Copyright 2001-2006 Paul James Mutton, http://www.jibble.org/
 *  Copyright 2007 Arnaud Blouin 
 *
 *  This file is part of jlibeps.
 * 
 *  jlibeps is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  jlibeps is distributed without any warranty; without even the 
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 *  PURPOSE. See the GNU General Public License for more details.<br>
 * 
 */

package org.sourceforge.jlibeps.epsgraphics;


/**
 * Copyright 2001-2006 Paul James Mutton, http://www.jibble.org/<br>
 * Copyright 2007 Arnaud Blouin<br>
 * 08/09/07
 * @version 0.1
 */
public class EpsException extends RuntimeException
{

	public EpsException(String message)
	{
		super(message);
	}

}
